from metamenth.enumerations import TriggerType
from metamenth.measure_instruments.interfaces.abstract_data_measure import AbstractDataMeasure


class TriggerHistory(AbstractDataMeasure):

    def __init__(self, trigger_type: TriggerType, value: float = None, timestamp: str = None):
        if value is None:
            value = 0.0
        super().__init__(value, timestamp)
        self.trigger_type = trigger_type

    def __eq__(self, other):
        if isinstance(other, TriggerHistory):
            return self.UID == other.UID
        return False



from metamenth.enumerations.abstract_enum import AbstractEnum


class HeatingType(AbstractEnum):
    """
    Types of heating

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    ELECTRIC = "Electric"
    HYDRONIC = "Hydronic"
    DIMPLEX = "Dimplex"
    OTHER = "Other"




from metamenth.energysystem.interfaces.abstract_electrical import AbstractElectrical
from metamenth.enumerations import PowerState
from metamenth.datatypes.binary_measure import BinaryMeasure


class Alternator(AbstractElectrical):
    def __init__(self, name: str, power_state: PowerState = PowerState.NONE):
        super().__init__(name, power_state)

        self._power_rating = None

    @property
    def power_rating(self) -> BinaryMeasure:
        return self._power_rating

    @power_rating.setter
    def power_rating(self, value: BinaryMeasure):
        self._power_rating = value

    def __str__(self):
        return (
            f"Alternator("
            f"{super().__str__()}"
            f"Power Rating: {self.power_rating})"
        )




from uuid import uuid4
from typing import List
from metamenth.datatypes.continuous_measure import ContinuousMeasure
from metamenth.datatypes.interfaces.abstract_dynamic_entity import AbstractDynamicEntity


class AbstractVentilationComponent(AbstractDynamicEntity):

    def __init__(self, name: str):
        super().__init__()
        self._UID = str(uuid4())
        self._name = None
        self._operating_conditions: List[ContinuousMeasure] = []

        self.name = name

    @property
    def UID(self) -> str:
        return self._UID

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, value: str):
        if value is not None:
            self._name = value
        else:
            raise ValueError("name must be of type str")

    @property
    def operating_conditions(self) -> [ContinuousMeasure]:
        return self._operating_conditions.copy() if self._operating_conditions else []

    @operating_conditions.setter
    def operating_conditions(self, value: [ContinuousMeasure]):
        if value is not None and type(value) is list:
            self._operating_conditions.extend(value)

    def __eq__(self, other):
        # subsystems are equal if they share the same name
        if isinstance(other, AbstractVentilationComponent):
            # Check for equality based on the 'name' attribute
            return self.name == other.name and self.UID == other.UID
        return False

    def __str__(self):
        return (
            f"UID: {self.UID}, "
            f"Name: {self.name}, "
            f"Rated Device Measure: {self.operating_conditions}"
        )


from enum import Enum


class MaterialType(Enum):
    """
    Different materials a layer in a building cover can be made of

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    EX_WALL_WOOD = "ExWall:Wood"
    EX_WALL_BRICK = "ExWall:Brick"
    EX_WALL_CONCRETE = "ExWall:Concrete"
    EX_WALL_STEEL = "ExWall:Steel"
    EX_WALL_ALUMINIUM = "ExWall:Aluminium"
    EX_WALL_FIRBRE_CEMENT = "ExWall:FibreCement"
    ROOF_STEEL = "Roof:Steel"
    ROOF_ALUMINIUM = "Roof:Aluminium"
    ROOF_COPPER = "Roof:Copper"
    ROOF_ZINC = "Roof:Zinc"
    ROOF_TIN = "Roof:Tin"
    ROOF_STAINLESS_STEEL = "Roof:StainlessSteel"
    ROOF_STONE_COATED_STEEL = "Roof:StoneCoatedSteel"
    ROOF_METAL_SHINGLES = "Roof:MetalShingles"
    ROOF_CONCRETE_TILE = "Roof:ConcreteTiles"
    ROOF_GREEN = "Roof:Green"
    ROOF_EPMD = "Roof:EPMD"
    WIN_DOOR_WOOD = "WindowDoor:Wood"
    WIN_DOOR_VINYL = "WindowDoor:Vinyl"
    WIN_DOOR_FIBRE_GLASS = "WindowDoor:FibreGlass"
    WIN_DOOR_ALUMINIUM = "WindowDoor:Aluminium"
    IN_WALL_FIRBRE_GLASS = "InWall:FibreGlass"
    IN_WALL_CELLULOSE = "InWall:Cellulose"
    IN_WALL_SPRAY_FOAM = "InWall:SprayFoam"
    IN_WALL_MINERAL_WOOL = "InWall:MineralWool"
    FLOOR_CONCRETE = "Floor:Concrete"
    FLOOR_Wood = "Floor:Wood"
    FLOOR_CERAMIC_TILE = "Floor:CeramicTile"
    FLOOR_NATURAL_STONE = "Floor:NaturalStone"
    FLOOR_VINYL = "Floor:Vinyl"
    FLOOR_BAMBOO = "Floor:Bamboo"
    FLOOR_CORK = "Floor:Cork"
    FLOOR_RUBBER = "Floor:Rubber"
    FLOOR_LINOLEUM = "Floor:Linoleum"



from metamenth.subsystem.hvac_components.interfaces.abstract_hvac_component import AbstractHVACComponent
from metamenth.subsystem.hvac_components.variable_frequency_drive import VariableFrequencyDrive


class AbstractDuctConnectedComponent(AbstractHVACComponent):

    def __init__(self, name: str):
        super().__init__(name)

        self._ducts: [] = []
        self._vfd = None

    @property
    def ducts(self) -> []:
        return self._ducts.copy()

    @property
    def vfd(self) -> VariableFrequencyDrive:
        return self._vfd

    def add_duct(self, duct):
        from metamenth.subsystem.hvac_components.duct import Duct
        if duct is not None and isinstance(duct, Duct):
            self._ducts.append(duct)
        else:
            raise ValueError("value provided is not a duct")

    @vfd.setter
    def vfd(self, value: VariableFrequencyDrive):
        self._vfd = value

    def __str__(self):
        return (
            f"({super().__str__()}"
            f"Ducts {self.ducts}, "
            f"VFD: {self.vfd}, "
        )


from metamenth.subsystem.hvac_components.interfaces.abstract_hvac_component import AbstractHVACComponent
from metamenth.enumerations import DamperType
from metamenth.measure_instruments.damper_position import DamperPosition
from typing import List
from typing import Dict
from metamenth.utils import StructureEntitySearch


class Damper(AbstractHVACComponent):
    def __init__(self, name: str, damper_type: DamperType):
        """
        Models a damper in hvac system
        :param name: the unique name of the heat exchanger
        :param damper_type: the type of damper
        """
        super().__init__(name)
        self._damper_type = None
        self._percentage_opened: [DamperPosition] = []

        self.damper_type = damper_type

    @property
    def damper_type(self) -> DamperType:
        return self._damper_type

    @damper_type.setter
    def damper_type(self, value: DamperType):
        if value is not None:
            self._damper_type = value
        else:
            raise ValueError("damper_type must be of type DamperType")

    def add_damper_position(self, damper_position: DamperPosition):
        return self._percentage_opened.append(damper_position)

    def remove_damper_position(self, damper_position: DamperPosition):
        self._percentage_opened.remove(damper_position)

    def get_damper_positions(self, search_terms: Dict = None) -> List[DamperPosition]:
        """
        Search data by attributes values
        :param search_terms: a dictionary of attributes and their values
        :return [DamperPosition]:
        """
        return StructureEntitySearch.search(self._percentage_opened, search_terms)

    def get_damper_positions_by_date(self, from_timestamp: str, to_timestamp: str = None) -> List[DamperPosition]:
        """
        searches damper positions data based on provided timestamp
        :param from_timestamp: the start timestamp
        :param to_timestamp: the end timestamp
        :return: [DamperPosition]
        """
        return StructureEntitySearch.date_range_search(self._percentage_opened, from_timestamp, to_timestamp)

    def __str__(self):
        return (
            f"Damper ({super().__str__()}"
            f"Type: {self.damper_type}, "
            f"Percentage Opened : {self._percentage_opened})"
        )



from metamenth.subsystem.hvac_components.interfaces.abstract_hvac_component import AbstractHVACComponent
from metamenth.enumerations import CoilMaterial
from metamenth.enumerations import RefrigerantType
from metamenth.datatypes.interfaces.abstract_measure import AbstractMeasure


class Condenser(AbstractHVACComponent):
    def __init__(self, name: str, refrigerant_type: RefrigerantType, coil_material: CoilMaterial,
                 heat_transfer_area: AbstractMeasure = None, heat_transfer_coefficient: float = 0.0):
        super().__init__(name)
        self._refrigerant_type = None
        self._coil_material = None
        self._heat_transfer_area = heat_transfer_area
        self._heat_transfer_coefficient = heat_transfer_coefficient

        self.refrigerant_type = refrigerant_type
        self.coil_material = coil_material

    @property
    def refrigerant_type(self) -> RefrigerantType:
        return self._refrigerant_type

    @refrigerant_type.setter
    def refrigerant_type(self, value: RefrigerantType):
        if not value:
            raise ValueError("refrigerant_type must be of type RefrigerantType")
        self._refrigerant_type = value

    @property
    def coil_material(self) -> CoilMaterial:
        return self._coil_material

    @coil_material.setter
    def coil_material(self, value: CoilMaterial):
        if not value:
            raise ValueError("coil_material must be of type CoilMaterial")
        self._coil_material = value

    @property
    def heat_transfer_area(self) -> AbstractMeasure:
        return self._heat_transfer_area

    @heat_transfer_area.setter
    def heat_transfer_area(self, value: AbstractMeasure):
        self._heat_transfer_area = value

    @property
    def heat_transfer_coefficient(self) -> float:
        return self._heat_transfer_coefficient

    @heat_transfer_coefficient.setter
    def heat_transfer_coefficient(self, value: float):
        self._heat_transfer_coefficient = value

    def calculate_heat_transfer_rate(self, temperature_difference: float) -> float:
        """
        computes the heat transfer rate of the condenser
        :param temperature_difference: the temperature difference
        :return:
        """
        if self._heat_transfer_area and self._heat_transfer_coefficient > 0.0:
            return self.heat_transfer_coefficient * self.heat_transfer_area.value * temperature_difference
        return 0.0



from metamenth.enumerations import MeasurementUnit
from metamenth.measure_instruments.interfaces.abstract_reader import AbstractReader
from metamenth.measure_instruments.electric_vehicle_connectivity import ElectricVehicleConnectivity
from metamenth.utils import EntityInsert
from metamenth.utils import StructureEntitySearch
from typing import Dict
from metamenth.enumerations import BuildingEntity


class EVChargingMeter(AbstractReader):
    """
    A representation of an electric vehicle charging meter

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """

    def __init__(self, meter_location: str, measurement_unit: MeasurementUnit,
                 manufacturer: str = None):
        """
        Initializes a Meter instance.

        :param meter_location: The what3word location of the meter.
        :param manufacturer: The manufacturer of the meter.
        :param measurement_unit: The measurement unit of the meter data.
        """
        super().__init__(measurement_unit, meter_location, manufacturer)
        self._vehicle_connectivity: [ElectricVehicleConnectivity] = []

    def get_connectivity_data(self, search_terms: Dict = None) -> [ElectricVehicleConnectivity]:
        """
        Search meter recordings by attributes values
        :param search_terms: a dictionary of attributes and their values
        :return [ElectricVehicleConnectivity]:
        """
        return StructureEntitySearch.search(self._vehicle_connectivity, search_terms)

    def get_connectivity_data_by_date(self, from_timestamp: str, to_timestamp: str = None) -> [ElectricVehicleConnectivity]:
        """
        searches meter recordings based on provided timestamp
        :param from_timestamp: the start timestamp
        :param to_timestamp: the end timestamp
        :return: [ElectricVehicleConnectivity]
        """
        return StructureEntitySearch.date_range_search(self._vehicle_connectivity, from_timestamp, to_timestamp)

    def add_meter_measure(self, connectivity_data: ElectricVehicleConnectivity):
        """
        Add vehicle connectivity for this meter
        :param connectivity_data: the recorded electric vehicle charging or discharging by the meter.
        """
        EntityInsert.insert_building_entity(self._vehicle_connectivity, connectivity_data, BuildingEntity.SCHEDULE.value)

    def __str__(self):
        """
        :return: A formatted string representing the meter.
        """

        return (
            f"EVChargingMeter("
            f"{super().__str__()})"
        )



from metamenth.visitors.interfaces.abstract_space_visitor import AbstractSpaceVisitor
from typing import Dict


class MeterSearchVisitor(AbstractSpaceVisitor):
    """
    A concrete visitor that searches for meters in
    building spaces or zones
    """

    def __init__(self, meter_criteria: Dict, floor_criteria: Dict = None,
                 room_criteria: Dict = None, open_space_criteria: Dict = None):
        """
        :param meter_criteria: the search criteria for meters
        """
        super().__init__(floor_criteria, room_criteria, open_space_criteria)
        self._meter_criteria = meter_criteria

    def visit_building(self, building):
        print(f'Visiting building: {building.address}')
        for meter in building.get_meters():
            if self._match_criteria(meter, self._meter_criteria):
                self.found_entities.append(meter)

        for floor in building.get_floors():
            floor.accept(self)

    def visit_room(self, room):
        if self._match_criteria(room, self._room_criteria):
            print(f'Visiting room: {room.name}')
            self._search_meters(room)

    def visit_open_space(self, open_space):
        if self._match_criteria(open_space, self._open_space_criteria):
            print(f'Visiting open space: {open_space.name}')
            self._search_meters(open_space)

    def _search_meters(self, space):
        if self._match_criteria(space.meter, self._meter_criteria):
            # compare meter in open space to search criteria
            if space.meter:
                self.found_entities.append(space.meter)

            # search HVAC components for meters
            self._search_entities(space.get_hvac_components())
            # search energy systems for meters
            self._search_entities(space.get_energy_systems())

    def _search_entities(self, entities):
        for entity in entities:
            if self._match_criteria(entity.meter, self._meter_criteria):
                if entity.meter:
                    self.found_entities.append(entity.meter)



from metamenth.enumerations.abstract_enum import AbstractEnum


class UPSPhase(AbstractEnum):
    """
    Different phases of uninterruptible power supply

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    SINGLE = "Single"
    THREE = "Three"
    SPLIT = "Split"



from metamenth.enumerations.abstract_enum import AbstractEnum


class ApplianceType(AbstractEnum):
    """
    Different types of appliances

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    LIGHT_BULB = "LightBulb"
    KETTLE = "Kettle"
    MICROWAVE = "Microwave"
    REFRIGERATOR = "Refrigerator"
    FREEZER = "Freezer"
    THERMOSTAT = "Thermostat"
    COFFEE_MAKER = "CoffeeMaker"
    DOOR_BELL = "DoorBell"
    LOCK = "Lock"
    OVEN = "Oven"
    AIR_CONDITIONER = "AirConditioner"
    PLUG = "Plug"
    CAMERA = "Camera"
    DISHWASHER = "Dishwasher"
    TOASTER = "Toaster"
    RICE_COOKER = "RiceCooker"
    WASHING_MACHINE = "WashingMachine"
    TELEVISION = "Television"
    DRYER = "Dryer"
    PRESSING_IRON = "PressingIron"
    ELECTRIC_FAN = "Electric"
    STOVE = "Stove"
    BLENDER = "Blender"
    PRINTER = "Printer"
    SCANNER = "Scanner"
    PAPER_SHREDDER = "PaperShredder"
    LAMINATOR = "Laminator"
    PHOTOCOPIER = "Photocopier"
    SMOKE_DETECTOR = "SmokeDetector"
    VACUUM_CLEANER = "VacuumCleaner"
    PROJECTOR = "Projector"
    INSTANT_POT = "InstantPot"
    LAPTOP = "Laptop"
    DESKTOP_COMPUTER = "DesktopComputer"
    SECURITY_CAMERA = "SecurityCamera"
    POWER_STRIP = "PowerStrip"
    SWITCH = "Switch"
    WHITEBOARD = "Whiteboard"
    NOTEBOOK = "Notebook"
    AIR_PURIFIER = "AirPurifier"
    HUMIDIFIER = "Humidifier"
    DEHUMIDIFIER = "Dehumidifier"
    SPEAKER = "Speaker"
    OTHER = "Other"




from enum import Enum


class DataMeasurementType(Enum):
    """
    Different weather data a building weather station can record.

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    OUTSIDE_TEMPERATURE = "OutsideTemperature"
    HUMIDITY = "Humidity"
    RELATIVE_HUMIDITY = "RelativeHumidity"
    PRESSURE = "Pressure"
    PRECIPITATION = "Precipitation"
    WIND_SPEED = "WindSpeed"
    WIND_DIRECTOR = "WindDirection"
    MOISTURE_CONTENT = "MoistureContent"
    SOLAR_RADIATION = "SolarRadiation"
    GLOBAL_NOMINAL_IRRADIANCE = "GlobalNominalIrradiance"
    DIFFUSE_HORIZONTAL_IRRADIANCE = "DiffuseHorizontalIrradiance"
    DIRECT_NOMINAL_IRRADIANCE = "DirectNominalIrradiance"
    GLOBAL_HORIZONTAL_IRRADIANCE = "GlobalHorizontalIrradiance"
    EXPORTED_ELECTRICITY = "ExportedElectricity"
    IMPORTED_ELECTRICITY = "ImportedElectricity"
    CONSUMED_ELECTRICITY = "ConsumedElectricity"



from metamenth.enumerations.abstract_enum import AbstractEnum


class SolarPVType(AbstractEnum):
    """
    Types of solar panels

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    BUILDING_INTEGRATED_PHOTOVOLTAIC = "BIPV"
    BUILDING_INTEGRATED_PHOTOVOLTAIC_THERMAL = "BIPVT"
    OTHER = "Other"



from metamenth.enumerations.abstract_enum import AbstractEnum


class VentilationType(AbstractEnum):
    """
    Different kinds of ventilation system

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    AIR_HANDLING_UNIT = "AirHandlingUnit"




from metamenth.enumerations.abstract_enum import AbstractEnum
from typing import Dict


class AbstractSpaceVisitor:
    """
    Defines an interface for visiting
    spaces in buildings
    """

    def __init__(self, floor_criteria: Dict,
                 room_criteria: Dict, open_space_criteria: Dict):
        """
        :param floor_criteria: criteria to filter down sensor search to specific floors
        :param room_criteria: criteria to filter down sensor search to specific rooms
        :param open_space_criteria: criteria to filter down sensor search to specific open spaces
        """

        self._floor_criteria = floor_criteria
        self._room_criteria = room_criteria
        self._open_space_criteria = open_space_criteria
        self.found_entities = []

    def visit_building(self, building):
        print(f'Visiting building: {building.address}')
        for floor in building.get_floors():
            floor.accept(self)

    def visit_floor(self, floor):
        if self._match_criteria(floor, self._floor_criteria):
            print(f'Visiting floor: {floor.number}')
            for room in floor.get_rooms():
                room.accept(self)

            for open_space in floor.get_open_spaces():
                open_space.accept(self)

    def visit_room(self, room):
        pass

    def visit_open_space(self, open_space):
        pass

    def _match_criteria(self, entity, criteria):
        """
        Searches for sensors that meet specific criteria
        :param entity: the entity being compare to search criteria
        :param criteria: the filter criteria
        """
        if not criteria:
            return True

        try:
            for key, value in criteria.items():
                if key == 'component_class':
                    continue
                att_value = entity.get(key)
                if isinstance(att_value, AbstractEnum):
                    att_value = att_value.value
                if isinstance(value, list):
                    # for list search criteria
                    if isinstance(att_value, list):
                        match_found = False
                        for i in range(len(att_value)):
                            for j in range(len(value)):
                                if att_value[i] == value[j]:
                                    match_found = True
                        return match_found

                    else:
                        if att_value not in value:
                            return False
                else:
                    # For single-value criteria
                    if att_value != value:
                        return False
        except AttributeError:
            # handle attribute errors for None types
            pass

        return True



from metamenth.enumerations.abstract_enum import AbstractEnum


class HeatExchangerType(AbstractEnum):
    """
    Types of heat exchangers

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    PLATE = "Plate"
    SHELL_AND_TUBE = "ShellAndTube"
    FIN_TUBE = "FinTube"
    PLATED_FIN = "PlatedFin"
    OTHER = "Other"




from metamenth.enumerations import MaterialType
from metamenth.datatypes.interfaces.abstract_measure import AbstractMeasure
import uuid
from metamenth.misc import Validate


class Material:
    """
    Material making up layers in the envelope of a building

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    def __init__(
        self,
        description: str,
        material_type: MaterialType,
        density: AbstractMeasure,
        heat_capacity: AbstractMeasure,
        thermal_transmittance: AbstractMeasure,
        thermal_resistance: AbstractMeasure,
        thermal_conductivity: AbstractMeasure = None,
        thermal_absorptance: float = None,
        solar_heat_gain_coefficient: float = None,
        solar_absorptance: float = None,
        visible_absorptance: float = None
    ):

        self._UID = str(uuid.uuid4())
        self._description = None
        self._material_type = None
        self._density = None
        self._heat_capacity = None
        self._thermal_transmittance = None
        self._thermal_resistance = None
        self._solar_heat_gain_coefficient = None
        self._solar_absportance = None
        self._thermal_absorptance = None
        self._visible_absorptance = None
        self._thermal_conductivity = thermal_conductivity

        self.description = description
        self.density = density
        self.material_type = material_type
        self.heat_capacity = heat_capacity
        self.thermal_resistance = thermal_resistance
        self.thermal_transmittance = thermal_transmittance
        self.solar_heat_gain_coefficient = solar_heat_gain_coefficient
        self.solar_absportance = solar_absorptance
        self.thermal_absorptance = thermal_absorptance
        self.visible_absorptance = visible_absorptance

    @property
    def UID(self) -> str:
        return self._UID

    @property
    def description(self) -> str:
        return self._description

    @description.setter
    def description(self, value: str):
        if value is None:
            raise ValueError("description must be a string")
        self._description = value

    @property
    def material_type(self) -> MaterialType:
        return self._material_type

    @material_type.setter
    def material_type(self, value: MaterialType):
        if value is None:
            raise ValueError("material_type must be of type MaterialType")
        self._material_type = value

    @property
    def density(self) -> AbstractMeasure:
        return self._density

    @density.setter
    def density(self, value: AbstractMeasure):
        if value is None:
            raise ValueError("density must be of type BinaryMeasure")
        self._density = value

    @property
    def heat_capacity(self) -> AbstractMeasure:
        return self._heat_capacity

    @heat_capacity.setter
    def heat_capacity(self, value: AbstractMeasure):
        if value is None:
            raise ValueError("heat_capacity must be of type BinaryMeasure")
        self._heat_capacity = value

    @property
    def thermal_transmittance(self) -> AbstractMeasure:
        return self._thermal_transmittance

    @thermal_transmittance.setter
    def thermal_transmittance(self, value: AbstractMeasure):
        if value is None:
            raise ValueError("thermal_transmittance must be of type BinaryMeasure")
        self._thermal_transmittance = value

    @property
    def thermal_conductivity(self) -> AbstractMeasure:
        return self._thermal_conductivity

    @thermal_conductivity.setter
    def thermal_conductivity(self, value: AbstractMeasure):
        self._thermal_conductivity = value

    @property
    def thermal_resistance(self) -> AbstractMeasure:
        return self._thermal_resistance

    @thermal_resistance.setter
    def thermal_resistance(self, value: AbstractMeasure):
        if value is None:
            raise ValueError("thermal_resistance must be of type BinaryMeasure")
        self._thermal_resistance = value

    @property
    def thermal_absorptance(self) -> float:
        return self._thermal_absorptance

    @thermal_absorptance.setter
    def thermal_absorptance(self, value: float):
        self._thermal_absorptance = Validate.validate_number_range(value, (0, 1))

    @property
    def solar_heat_gain_coefficient(self) -> float:
        return self._solar_heat_gain_coefficient

    @solar_heat_gain_coefficient.setter
    def solar_heat_gain_coefficient(self, value: float):
        self._solar_heat_gain_coefficient = Validate.validate_number_range(value, (0, 1))

    @property
    def solar_absorptance(self) -> float:
        return self._solar_absportance

    @solar_absorptance.setter
    def solar_absorptance(self, value: float):
        self._solar_absportance = Validate.validate_number_range(value, (0, 1))

    @property
    def visible_absorptance(self) -> float:
        return self._visible_absorptance

    @visible_absorptance.setter
    def visible_absorptance(self, value: float):
        self._visible_absportance = Validate.validate_number_range(value, (0, 1))

    def __str__(self):
        return (
            f"Material("
            f"UID: {self.UID}, "
            f"Description: {self.description}, "
            f"Type: {self.material_type.value}, "
            f"Density: {self.density.value} {self.density.measurement_unit.value}, "
            f"Heat Capacity: {self.heat_capacity.value} {self.heat_capacity.measurement_unit.value}, "
            f"Thermal Transmittance: {self.thermal_transmittance.value} {self.thermal_transmittance.measurement_unit.value}, "
            f"Thermal Resistance: {self.thermal_resistance.value} {self.thermal_resistance.measurement_unit.value}, "
            f"Thermal Conductivity: "
            f"{self.thermal_conductivity.value if self.thermal_conductivity is not None else None} "
            f"{self.thermal_conductivity.measurement_unit.value if self.thermal_conductivity is not None else ''}, "
            f"Thermal Absorptance: {self.thermal_absorptance}, "
            f"Solar Heat Gain Coefficient: {self.solar_heat_gain_coefficient}, "
            f"Solar Absorptance: {self.solar_absorptance}, "
            f"Visible Absorptance: {self.visible_absorptance}"
        )




from metamenth.enumerations.abstract_enum import AbstractEnum


class ATSPowerSourceType(AbstractEnum):
    """
    Type of ATS power sources

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    UTILITY_GENERATOR = "UtilityGenerator"
    UTILITY_UTILITY = "UtilityUtility"
    GENERATOR_GENERATOR = "GeneratorGenerator"
    UTILITY_UTILITY_GENERATOR = "UtilityUtilityGenerator"
    UTILITY_GENERATOR_GENERATOR = "UtilityGeneratorGenerator"
    OTHER = "Other"



from metamenth.enumerations.abstract_enum import AbstractEnum


class HeatExchangerFlowType(AbstractEnum):
    """
    Heat exchanger flow types

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    PARALLEL = "Parallel"
    COUNTER = "Counter"
    OTHER = "Other"




from metamenth.energysystem.interfaces.abstract_electrical import AbstractElectrical
from metamenth.enumerations import PowerState
from metamenth.enumerations import UPSPhase
from metamenth.datatypes.binary_measure import BinaryMeasure
from metamenth.energysystem.storage_system.immobile_storage_energy_system import ImmobileStorageEnergySystem
from metamenth.enumerations import BuildingEntity
from metamenth.utils import EntityInsert
from metamenth.utils import EntityRemover
from metamenth.utils import StructureEntitySearch


class UninterruptiblePowerSupply(AbstractElectrical):
    def __init__(self, name: str, power_state: PowerState, phase: UPSPhase):
        super().__init__(name, power_state)
        self._phase = None
        self._noise_filtering = False
        self._surge_suppression = False
        self._power_rating = None
        self._storage_systems: [] = [ImmobileStorageEnergySystem]

        self.phase = phase

    @property
    def phase(self) -> UPSPhase:
        return self._phase

    @phase.setter
    def phase(self, value: UPSPhase):
        if value is None:
            raise ValueError('phase should be of type UPSPhase')
        self._phase = value

    @property
    def noise_filtering(self) -> bool:
        return self._noise_filtering

    @noise_filtering.setter
    def noise_filtering(self, value: bool):
        self._noise_filtering = value

    @property
    def surge_suppression(self) -> bool:
        return self._surge_suppression

    @surge_suppression.setter
    def surge_suppression(self, value: bool):
        self._surge_suppression = value

    @property
    def power_rating(self) -> BinaryMeasure:
        return self._power_rating

    @power_rating.setter
    def power_rating(self, value: BinaryMeasure):
        self._power_rating = value

    def add_storage_system(self, storage_system: ImmobileStorageEnergySystem):
        """
        adds storage system to UPS
        :param storage_system: the storage system to add
        :return:
        """
        EntityInsert.insert_building_entity(self._storage_systems, storage_system, BuildingEntity.HVAC_COMPONENT.value)

    def remove_storage_system(self, storage_system: ImmobileStorageEnergySystem):
        """
        Removes storage system from UPS
        :param storage_system: the storage system to remove
        :return:
        """
        EntityRemover.remove_building_entity(self._storage_systems, storage_system)

    def get_storage_system_by_name(self, name) -> ImmobileStorageEnergySystem:
        """
        Search storage system by name
        :param name: the name of the storage system
        :return:
        """
        return StructureEntitySearch.search_by_name(self._storage_systems, name)

    def __str__(self):
        return (
            f"UninterruptiblePowerSupply("
            f"{super().__str__()}"
            f"Phase: {self.phase}, "
            f"Noise Filtering: {self.noise_filtering}, "
            f"Surge Suppression: {self.surge_suppression}, "
            f"Power Rating: {self.power_rating})"
        )



from metamenth.subsystem.hvac_components.interfaces.abstract_duct_connected_component import AbstractDuctConnectedComponent


class CoolingTower(AbstractDuctConnectedComponent):
    def __init__(self, name: str):
        """
        Models a cooling tower in an hvac system
        :param name: the unique name of the boiler
        :
        """
        super().__init__(name)

    def __str__(self):
        return (
            f"Cooling Tower ({super().__str__()})"
        )



from datetime import datetime
from uuid import uuid4
from metamenth.datatypes.interfaces.abstract_measure import AbstractMeasure
from metamenth.misc import Validate


class WeatherData:
    def __init__(self, data: AbstractMeasure, timestamp: str = None):
        """
        :param data: The binary measure (value and unit) of the weather data.

        """
        self._UID = str(uuid4())  # Generating a unique identifier
        self._timestamp = datetime.now().replace(microsecond=0) if timestamp is None else Validate.parse_date(timestamp)
        self._data = None

        self.data = data

    @property
    def UID(self) -> str:
        return self._UID

    @property
    def data(self) -> AbstractMeasure:
        return self._data

    @data.setter
    def data(self, value: AbstractMeasure):
        self._data = value

    @property
    def timestamp(self) -> datetime:
        return self._timestamp

    def __str__(self):
        """
        Returns a string representation of the BuildingWeatherData instance.

        :return: A formatted string representing the BuildingWeatherData details.
        """
        return (
            f"BuildingWeatherData("
            f"UID: {self.UID}, "
            f"Timestamp: {self.timestamp}, "
            f"Data: {self.data})"
        )



from enum import Enum


class TriggerType(Enum):
    """
    How different actuators can be triggered

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    ON = "On"
    OFF = "Off"
    OPEN = "Open"
    CLOSE = "Close"
    OPEN_CLOSE = "OpenClose"





from metamenth.subsystem.hvac_components.interfaces.abstract_duct_connected_component import AbstractDuctConnectedComponent
from metamenth.enumerations import HeatExchangerType
from metamenth.enumerations import HeatExchangerFlowType


class HeatExchanger(AbstractDuctConnectedComponent):
    def __init__(self, name: str, heat_exchanger_type: HeatExchangerType,
                 heat_exchanger_flow_type: HeatExchangerFlowType):
        """
        Models a heat exchanger in a built environment
        :param name: the unique name of the heat exchanger
        :param heat_exchanger_type: the type of heat exchanger
        :param heat_exchanger_flow_type: the substance flow type of the heat exchanger
        """
        super().__init__(name)
        self._heat_exchanger_type = None
        self._heat_exchanger_flow_type = None

        self.heat_exchanger_type = heat_exchanger_type
        self.heat_exchanger_flow_type = heat_exchanger_flow_type

    @property
    def heat_exchanger_type(self) -> HeatExchangerType:
        return self._heat_exchanger_type

    @heat_exchanger_type.setter
    def heat_exchanger_type(self, value: HeatExchangerType):
        if value is not None:
            self._heat_exchanger_type = value
        else:
            raise ValueError("heat_exchanger_type must be of type HeatExchangerType")

    @property
    def heat_exchanger_flow_type(self) -> HeatExchangerFlowType:
        return self._heat_exchanger_flow_type

    @heat_exchanger_flow_type.setter
    def heat_exchanger_flow_type(self, value: HeatExchangerFlowType):
        if value is not None:
            self._heat_exchanger_flow_type = value
        else:
            raise ValueError("heat_exchanger_flow_type must be of type HeatExchangerFlowType")

    def __str__(self):
        return (
            f"HeatExchanger ({super().__str__()}"
            f"Type: {self.heat_exchanger_type.value}, "
            f"Flow Type : {self.heat_exchanger_flow_type.value})"
        )



import uuid
from metamenth.datatypes.interfaces.abstract_measure import AbstractMeasure
from metamenth.misc import Validate
from metamenth.virtual.interfaces.abstract_zonal_entity import AbstractZonalEntity
from metamenth.datatypes.schedulable_entity import SchedulableEntity
from metamenth.structure.envelope import Envelope


class AbstractSpace(AbstractZonalEntity):
    """
    An abstract class for spaces in a building
    """

    def __init__(self, area: AbstractMeasure, location: str = None):
        """
        :param area: The area of the space.
        :param location: The location of the space (three words delimited with two periods).
        """
        super().__init__()
        self._UID = str(uuid.uuid4())
        self._area = None
        self._location = None
        self._schedulable_entity = SchedulableEntity()
        self._envelope = None

        # Apply validation
        self.area = area
        self.location = location

    @property
    def UID(self) -> str:
        return self._UID

    @property
    def area(self) -> AbstractMeasure:
        return self._area

    @area.setter
    def area(self, value: AbstractMeasure):
        if value is None:
            raise ValueError('area must be of type BinaryMeasure')
        self._area = value

    @property
    def location(self) -> str:
        return self._location

    @location.setter
    def location(self, value: str):
        self._location = Validate.validate_what3word(value)

    @property
    def schedulable_entity(self) -> SchedulableEntity:
        return self._schedulable_entity

    @schedulable_entity.setter
    def schedulable_entity(self, value: SchedulableEntity):
        if value is None:
            raise ValueError("schedules should be of type SchedulableEntity")
        self._schedulable_entity = value

    @property
    def envelope(self) -> Envelope:
        return self._envelope

    @envelope.setter
    def envelope(self, value: Envelope):
        self._envelope = value

    def accept(self, visitor):
        pass

    def get(self, attribute):
        return getattr(self, attribute, None)

    def __str__(self):
        return (
            f"UID: {self.UID}, "
            f"Area: {self.area}, "
            f"Location: {self.location}, "
            f"Zones: {self._zones}, "
            f"Envelope: {self._envelope}, "
            f"Operational Schedule: {self._schedulable_entity}, "
        )



from metamenth.subsystem.hvac_components.interfaces.abstract_hvac_component import AbstractHVACComponent
from metamenth.enumerations import PowerState

from metamenth.subsystem.hvac_components.variable_frequency_drive import VariableFrequencyDrive


class Fan(AbstractHVACComponent):
    def __init__(self, name: str, power_state: PowerState, vfd: VariableFrequencyDrive = None):
        """
        Models a fan in an hvac system
        :param name: the unique name of the heat exchanger
        :param power_state: the power state of the fan
        """
        super().__init__(name)
        self._power_state = None
        self._vfd = vfd
        self.power_state = power_state

    @property
    def power_state(self) -> PowerState:
        return self._power_state

    @power_state.setter
    def power_state(self, value: PowerState):
        if value is not None:
            self._power_state = value
        else:
            raise ValueError("power_state must be of type PowerState")

    @property
    def vfd(self) -> VariableFrequencyDrive:
        return self._vfd

    @vfd.setter
    def vfd(self, value: VariableFrequencyDrive):
        self._vfd = value

    def __str__(self):
        return (
            f"Fan ({super().__str__()}"
            f"VFD: {self.vfd}, "
            f"Power State: {self.power_state})"
        )



from metamenth.enumerations.abstract_enum import AbstractEnum


class SolarDistributionType(AbstractEnum):
    """
    Solar distribution of buildings

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    FULL_EXTERIOR = "FullExterior"
    MINIMAL_SHADOWING = "MinimalShadowing"
    FULL_INTERIOR_AND_EXTERIOR = "FullInteriorAndExterior"
    FULL_EXTERIOR_WITH_REFLECTIONS = "FullExteriorWithReflections"
    FULL_INTERIOR_AND_EXTERIOR_WITH_REFLECTIONS = "FullInteriorAndExteriorWithReflections"



from metamenth.energysystem.storage_system.mobile_storage_energy_system import MobileStorageEnergySystem
from metamenth.enumerations import MeasurementUnit
from metamenth.datatypes.binary_measure import BinaryMeasure
from metamenth.enumerations import V2GMode


class ElectricVehicle(MobileStorageEnergySystem):

    def __init__(self, name: str, inverter: bool, unit: MeasurementUnit, v2g_capability: bool = False):
        super().__init__(name, inverter, unit)
        self._v2g_capability = v2g_capability
        self._v2g_power_limit = None
        self._v2g_mode = None

    @property
    def v2g_capability(self) -> bool:
        return self._v2g_capability

    @v2g_capability.setter
    def v2g_capability(self, value: bool):
        self._v2g_capability = value

    @property
    def v2g_power_limit(self) -> BinaryMeasure:
        return self._v2g_power_limit

    @v2g_power_limit.setter
    def v2g_power_limit(self, value: BinaryMeasure):
        self._v2g_power_limit = value

    @property
    def v2g_mode(self) -> V2GMode:
        return self._v2g_mode

    @v2g_mode.setter
    def v2g_mode(self, value: V2GMode):
        self._v2g_mode = value

    def __str__(self):
        return (
            f"ElectricVehicle("
            f"{super().__str__()} "
            f"V2G Capability: {self.v2g_capability}, "
            f"V2G Power Limit: {self.v2g_power_limit}, "
            f"V2G Mode: {self.v2g_mode})"
        )


from metamenth.enumerations.abstract_enum import AbstractEnum


class BatteryTech(AbstractEnum):
    """
    Battery technologies

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    LITHIUM = "Lithium"
    RUTHENIUM = "Ruthenium"
    NICKEL = "Nickel"
    OTHER = "Other"



from metamenth.datatypes.interfaces.abstract_range_measure import AbstractRangeMeasure
from metamenth.datatypes.measure import Measure


class ContinuousMeasure(AbstractRangeMeasure):
    """
    Continuous measurement

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """

    def __init__(self, measure: Measure):
        super().__init__(measure)

    def __str__(self):
        return (
            f"ContinuousMeasure("
            f"{super().__str__()})"
        )



import uuid
from metamenth.enumerations import MeasurementUnit
from metamenth.misc import Validate
from abc import ABC


class AbstractReader(ABC):
    """
    Abstract representation of meters

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """

    def __init__(self, measurement_unit: MeasurementUnit, meter_location: str, manufacturer: str = None):
        """
        Initializes a reader instance.

        :param meter_location: The what3word location of the meter.
        """
        self._UID = str(uuid.uuid4())
        self._meter_location = Validate.validate_what3word(meter_location)
        self._measurement_unit = None
        self._manufacturer = manufacturer

        # apply validation
        self.measurement_unit = measurement_unit

    @property
    def UID(self):
        return self._UID

    @property
    def meter_location(self):
        return self._meter_location

    @meter_location.setter
    def meter_location(self, value):
        self._meter_location = Validate.validate_what3word(value)

    @property
    def measurement_unit(self) -> MeasurementUnit:
        return self._measurement_unit

    @measurement_unit.setter
    def measurement_unit(self, value: MeasurementUnit):
        if value is not None:
            self._measurement_unit = value
        else:
            raise ValueError("measurement_unit must be of type MeasurementUnit")

    @property
    def manufacturer(self) -> str:
        return self._manufacturer

    @manufacturer.setter
    def manufacturer(self, value: str):
        self._manufacturer = value

    def get(self, attribute):
        return getattr(self, attribute, None)

    def __eq__(self, other):
        # Meters are equal if they share the same UID
        if isinstance(other, AbstractReader):
            # Check for equality based on the 'UID' attribute
            return self.UID == other.UID and self.measurement_unit == other.measurement_unit
        return False

    def __str__(self):
        """
        :return: A formatted string representing the recorder.
        """

        return (
            f"UID: {self.UID}, "
            f"Measurement Unit: {self.measurement_unit}, "
            f"Manufacturer: {self.manufacturer}"
        )



from metamenth.subsystem.interfaces.abstract_ventilation_component import AbstractVentilationComponent


class ThermalStorage(AbstractVentilationComponent):

    def __init__(self, name: str):
        super().__init__(name)

    def __str__(self):
        return (
            f"ThermalStorage ({super().__str__()})"
        )




from metamenth.enumerations.abstract_enum import AbstractEnum


class PumpType(AbstractEnum):
    """
    Different types of pumps (not heat pumps)

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    CENTRIFUGAL = "On"
    CIRCULATOR = "Off"
    CONDENSER = "StandBy"
    BOOSTER = "OutOfService"
    CONDENSATE = "VariableSpeed"
    OTHER = "Other"




from .abstract_measure import AbstractMeasure
from metamenth.datatypes.measure import Measure


class AbstractRangeMeasure(AbstractMeasure):
    """
    Defines properties shared by linear, continuous and exponential measures

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """

    def __init__(self, measure: Measure):
        super().__init__(measure)
        self.minimum = measure.minimum
        self.maximum = measure.maximum

    def __str__(self):
        return (
            f"Minimum: {self.minimum}, "
            f"Maximum: {self.maximum}, "
            f"{super().__str__()}"
        )



from setuptools import setup, find_packages

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setup(
    name="metamenth",
    version="1.0.5",
    packages=find_packages(include=["metamenth", "metamenth.*"]),
    install_requires=requirements,
    entry_points={
        'console_scripts': [
        ],
    },
    author="Peter Yefi",
    author_email="peteryefi@gmail.com",
    description="Metamodel for Energy Things (MetamEnTh) is an object-oriented metamodel designed to model the "
                "operational aspects of buildings. It focuses on the relationships and interactions between "
                "mechanical, electrical, and plumbing (MEP) entities and their connections to sensors and spatial "
                "entities such as rooms and open spaces within buildings.",
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url="https://github.com/ptidejteam/metamenth",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    license="Apache License 2.0",
)



from metamenth.enumerations.abstract_enum import AbstractEnum


class AirVolumeType(AbstractEnum):
    """
    Types of air volume boxes

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    VARIABLE_AIR_VOLUME = "VariableAirVolume"
    CONSTANT_AIR_VOLUME = "ConstantAirVolume"




from typing import Dict
import sys
from datetime import datetime
from typing import Union
from typing import List
from metamenth.measure_instruments.sensor_data import SensorData
from metamenth.measure_instruments.trigger_history import TriggerHistory
from metamenth.measure_instruments.meter_measure import MeterMeasure
from metamenth.measure_instruments.weather_data import WeatherData
from metamenth.misc import Validate
from metamenth.enumerations.abstract_enum import AbstractEnum


class StructureEntitySearch:
    """
    A visitor that entities in structures, e.g., meter, weather, stations, etc
    """

    def __init__(self):
        pass

    @staticmethod
    def search_by_id(entity_list, uid):
        """
        search structures by unique identifiers
        :param entity_list: the list of entity to search for a particular entity
        :param uid: the unique identifiers
        :return:
        """
        return StructureEntitySearch.search_structure_entity(entity_list, 'UID', uid)

    @staticmethod
    def search_by_name(entity_list, name):
        """
        search structures by name
        :param entity_list: the list of entity to search for a particular entity
        :param name: name of the structure
        :return:
        """
        return StructureEntitySearch.search_structure_entity(entity_list, 'name', name)

    @staticmethod
    def search(entity_list, search_terms: Dict):
        """
        search entities based on attribute values
        :param entity_list: the list of entity to search for a particular entity
        :param search_terms: key value pair of attributes and their values
        :return:
        """
        results = []
        if search_terms is None:
            return entity_list

        for entity in entity_list:
            found = True
            try:
                for attribute, value in search_terms.items():
                    att_value = getattr(entity, attribute)
                    if isinstance(att_value, AbstractEnum):
                        att_value = att_value.value
                    if att_value != value:
                        found = False
                if found:
                    results.append(entity)
            except AttributeError as err:
                # TODO: log errors to file
                print(err, file=sys.stderr)

        return results

    @staticmethod
    def date_range_search(entity_list: Union[List[SensorData], List[TriggerHistory], List[MeterMeasure],
                                             List[WeatherData]], from_timestamp: str, to_timestamp: str = None):
        """

        :param entity_list: a list of sensor, actuator or meter data
        :param from_timestamp: the start timestamp
        :param to_timestamp: the end timestamp
        :return:
        """
        if len(from_timestamp) == 10:  # Check if only date is provided
            from_timestamp += ' 00:00:00'  # Add default time of midnight

        if to_timestamp is None:
            to_tp = datetime.now().replace(microsecond=0)
        else:
            if len(to_timestamp) == 10:  # Check if only date is provided
                to_timestamp += ' 23:59:59'
            to_tp = Validate.parse_date(to_timestamp)
        from_tp = Validate.parse_date(from_timestamp)
        filtered_data = []
        for data in entity_list:
            if from_tp <= data.timestamp <= to_tp:
                filtered_data.append(data)

        return filtered_data

    @staticmethod
    def search_structure_entity(entity_list, search_field, search_value):
        """
        Search for structure floors, rooms, open spaces in a building
        :param entity_list: the list of entities to search
        :param search_field: the search field
        :param search_value: the search value
        :return:
        """

        for entity in entity_list:
            try:
                if getattr(entity, search_field) == search_value:
                    return entity
            except AttributeError as err:
                # TODO: log errors to file
                print(err, file=sys.stderr)
        return None



from metamenth.enumerations.abstract_enum import AbstractEnum


class CirculationPumpType(AbstractEnum):
    """
    Types of circulation pumps

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    CENTRIFUGAL = "Centrifugal"
    CONSTANT_PRESSURE = "ConstantPressure"
    OTHER = "Other"




from metamenth.structure.interfaces.abstract_space import AbstractSpace
from metamenth.enumerations import DuctConnectionEntityType
from metamenth.subsystem.hvac_components.circulation_pump import CirculationPump
from metamenth.subsystem.hvac_components.compressor import Compressor
from metamenth.subsystem.hvac_components.pump import Pump
from metamenth.subsystem.hvac_components.heat_pump import HeatPump
from metamenth.subsystem.hvac_components.condenser import Condenser
from metamenth.subsystem.hvac_components.heat_exchanger import HeatExchanger
from metamenth.subsystem.hvac_components.air_volume_box import AirVolumeBox
from metamenth.subsystem.hvac_components.cooling_tower import CoolingTower
from metamenth.subsystem.hvac_components.chiller import Chiller
from metamenth.subsystem.hvac_components.boiler import Boiler
from metamenth.utils import StructureEntitySearch
from typing import Dict
from metamenth.subsystem.hvac_components.fan_coil_unit import FanCoilUnit


class DuctConnection:
    def __init__(self):
        self._source_entities = []
        self._destination_entities = []
        self._is_loop = False

    def add_entity(self, entity_type: DuctConnectionEntityType, duct_entity):
        """
        Adds an entity to the duct connection
        :param entity_type: the type of duct entity: source, destination, inside
        :param duct_entity: the entity to add
        :return:
        """
        from metamenth.subsystem.hvac_components.duct import Duct
        allowed_entity_types = [
            HeatExchanger, AbstractSpace, Duct, Boiler, Chiller, CirculationPump,
            Compressor, Pump, HeatPump, Condenser, AirVolumeBox, CoolingTower, FanCoilUnit
        ]

        if not any(isinstance(duct_entity, cls) for cls in allowed_entity_types):
            raise ValueError(f'{duct_entity.name} cannot be connected to a duct')

        if isinstance(duct_entity, FanCoilUnit) and not duct_entity.is_ducted:
            raise ValueError(f'{duct_entity.name} cannot be connected to a duct')

        if entity_type == DuctConnectionEntityType.SOURCE:
            if duct_entity not in self._destination_entities:
                self._source_entities.append(duct_entity)
        elif entity_type == DuctConnectionEntityType.DESTINATION:
            if duct_entity not in self._source_entities:
                self._destination_entities.append(duct_entity)

    def remove_entity(self, entity_type: DuctConnectionEntityType, duct_entity):
        """
        Removes an entity from the duct connection
        :param entity_type: the type of duct entity: source, destination, inside
        :param duct_entity: the entity to add
        :return:
        """
        if entity_type == DuctConnectionEntityType.SOURCE:
            if duct_entity in self._source_entities:
                self._source_entities.remove(duct_entity)
        elif entity_type == DuctConnectionEntityType.DESTINATION:
            if duct_entity in self._destination_entities:
                self._destination_entities.remove(duct_entity)

    def get_source_entities(self, search_terms: Dict = None):
        """
        Search source entities by attribute values
        :param search_terms: a dictionary of attributes and their values
        :return:
        """
        return StructureEntitySearch.search(self._source_entities, search_terms)

    def get_destination_entities(self, search_terms: Dict = None):
        """
        Search destination entities by attribute values
        :param search_terms: a dictionary of attributes and their values
        :return:
        """
        return StructureEntitySearch.search(self._destination_entities, search_terms)

    @property
    def is_loop(self) -> bool:
        return self._is_loop

    @is_loop.setter
    def is_loop(self, value: bool):
        if value:
            if len(self._source_entities) <= 1 and len(self._destination_entities) <= 1:
                self._is_loop = value
            else:
                raise ValueError("is_loop can only be true if there are single source and destination entities")
        else:
            self._is_loop = value

    def __str__(self):
        return (
            f"DuctConnection ({super().__str__()}"
            f"Source Entities: {self._source_entities}, "
            f"Destination Entities: {self._destination_entities}, "
            f"Is Loop: {self._is_loop})"
        )




from metamenth.enumerations.abstract_enum import AbstractEnum


class V2GMode(AbstractEnum):
    """
    Vehicle to grid mode of electric vehicle

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    FREQUENCY_REGULATION = "FrequencyRegulation"
    PEAK_SHAVING = "PeakShaving"
    ENERGY_ARBITRAGE = "EnergyArbitrage"
    OTHER = "Other"



from datetime import datetime
from dataclasses import dataclass
from typing import Type
from metamenth.datatypes.interfaces.abstract_measure import AbstractMeasure
from uuid import uuid4


@dataclass
class OperationalSchedule:
    UID = uuid4()
    name: str # unique name for operational schedule
    start_date: datetime
    end_date: datetime
    setPoint: Type[AbstractMeasure] = None
    recurring: bool = True

    def __eq__(self, other):
        # schedules are equal if they share the same name
        if isinstance(other, OperationalSchedule):
            # Check for equality based on the 'name' attribute
            return self.name == other.name
        return False



from metamenth.enumerations.abstract_enum import AbstractEnum


class FloorType(AbstractEnum):
    """
    Different types of floors in a building.

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    REGULAR = "Regular"
    BASEMENT = "Basement"
    ROOFTOP = "Rooftop"



from metamenth.enumerations.abstract_enum import AbstractEnum


class EngineMode(AbstractEnum):
    """
    Modes of Engine

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    HEATING = "Heating"
    ELECTRICITY = "Electricity"
    CO_GENERATION = "CoGeneration"
    FUEL = "Fuel"
    OTHER = "Other"



from abc import ABC
from abc import abstractmethod
from metamenth.datatypes.observable_message import ObservableMessage


# Observer interface
class Observer(ABC):
    """
    An interface that defines method to log state of entities
    that need their state to be tracked
    """
    @abstractmethod
    def log_state(self, message: ObservableMessage):
        """
        logs the state of an object
        :param message: object {entity_type, entity_id, state, message}
        :return:
        """
        pass



from typing import List
from metamenth.datatypes.operational_schedule import OperationalSchedule
from metamenth.utils import EntityRemover
from metamenth.utils import EntityInsert
from metamenth.utils import StructureEntitySearch
from typing import Dict
from metamenth.enumerations import BuildingEntity


class SchedulableEntity:
    def __init__(self):
        self._schedules: List[OperationalSchedule] = []

    def add_schedule(self, schedule: OperationalSchedule):
        """
        Adds an operational schedule to this building
        :param schedule: the schedule
        :return:
        """
        EntityInsert.insert_building_entity(self._schedules, schedule, BuildingEntity.SCHEDULE.value)
        return self

    def remove_schedule(self, schedule: OperationalSchedule):
        """
        Removes an operational schedule from this building
        :param schedule: the schedule
        :return:
        """
        EntityRemover.remove_building_entity(self._schedules, schedule)

    def get_schedule_by_name(self, name) -> OperationalSchedule:
        """
        Search schedules by name
        :param name: the name of the schedule
        :return:
        """
        return StructureEntitySearch.search_by_name(self._schedules, name)

    def get_schedule_by_uid(self, uid) -> OperationalSchedule:
        """
        Search schedule by uid
        :param uid: the unique identifier of the schedule
        :return:
        """
        return StructureEntitySearch.search_by_id(self._schedules, uid)

    def get_schedules(self, search_terms: Dict = None) -> [OperationalSchedule]:
        """
        Search schedules by attributes values
        :param search_terms: a dictionary of attributes and their values
        :return:
        """
        return StructureEntitySearch.search(self._schedules, search_terms)




from metamenth.enumerations.abstract_enum import AbstractEnum


class TerrainType(AbstractEnum):
    """
   Terrain types of buildings

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    SUBURBS = "Suburbs"
    COUNTRY = "Country"
    CITY = "City"
    OCEAN = "Ocean"
    URBAN = "Urban"




from metamenth.enumerations.abstract_enum import AbstractEnum


class FilterType(AbstractEnum):
    """
    Different types of HVAC filters

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """

    FIBREGLASS = "FibreGlass"
    PLEATED = "Pleated"
    HIGH_EFFICIENCY_PARTICULAR_AIR = "HighEfficiencyParticularAir"
    ELECTROSTATIC = "Electrostatic"
    CARBON = "Carbon"
    ULTRAVIOLET_GERMICIDAL_IRRADIATION = "UltravioletGermicidalIrradiation"
    MEDIA = "Media"
    OTHER = "Other"




from metamenth.subsystem.hvac_components.interfaces.abstract_duct_connected_component import AbstractDuctConnectedComponent
from metamenth.enumerations import PumpType
from metamenth.datatypes.binary_measure import BinaryMeasure
from metamenth.misc.validate import Validate


class Pump(AbstractDuctConnectedComponent):
    def __init__(self, name: str, pump_type: PumpType,
                 flow_rate: BinaryMeasure = None,
                 power_consumption: BinaryMeasure = None,
                 efficiency: float = None):
        """
        Models a pump in a built environment
        :param name: the unique name of pump
        :param pump_type: the type of pump
        :param flow_rate: the flow rate of the pump mostly measured in liters per second (L/s)
        :param power_consumption: the power consumption of the pump in kilowatts (kW)
        :param efficiency: the efficiency of the pump (as a percentage, between 0 and 100)
        """
        super().__init__(name)
        self._pump_type = None
        self._power_consumption = power_consumption
        self._flow_rate = None
        self._efficiency = None

        self.flow_rate = flow_rate
        self.efficiency = efficiency
        self.pump_type = pump_type

    @property
    def pump_type(self) -> PumpType:
        return self._pump_type

    @pump_type.setter
    def pump_type(self, value: PumpType):
        if value is not None:
            self._pump_type = value
        else:
            raise ValueError("heat_exchanger_type must be of type HeatExchangerType")

    @property
    def power_consumption(self) -> BinaryMeasure:
        return self._power_consumption

    @power_consumption.setter
    def power_consumption(self, value: BinaryMeasure):
        self._power_consumption = value

    @property
    def flow_rate(self) -> BinaryMeasure:
        return self._flow_rate

    @flow_rate.setter
    def flow_rate(self, data: BinaryMeasure):
        if data and data.value > 0:
            self._flow_rate = data

    @property
    def efficiency(self) -> float:
        return self._efficiency

    @efficiency.setter
    def efficiency(self, value: float):
        self._efficiency = Validate.validate_number_range(value, (0, 1))


    def __str__(self):
        return (
            f"Pump ({super().__str__()}"
            f"Type: {self.pump_type.value}, "
            f"Flow Rate : {self.flow_rate.value if self.flow_rate else None}, "
            f"{self.flow_rate.measurement_unit if self.flow_rate else None}, "
            f"Efficiency: {self.efficiency}, "
            f"Power Consumption: {self.power_consumption.value if self.power_consumption else None}, "
            f"{self.power_consumption.measurement_unit if self.power_consumption else None})"
        )



from metamenth.enumerations import ZoneType
from metamenth.enumerations import HVACType
from uuid import uuid4
from typing import List
from metamenth.structure.interfaces.abstract_space import AbstractSpace
from metamenth.utils import EntityRemover
from metamenth.enumerations import BuildingEntity
from metamenth.utils import EntityInsert
from metamenth.utils import StructureEntitySearch
from metamenth.utils import StructureSearch
from typing import Dict


class Zone:
    """
    A zone in a building e.g. HVAC (thermal) zone

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """

    def __init__(self,
                 name: str,
                 zone_type: ZoneType,
                 hvac_type: HVACType = HVACType.NONE,
                 description: str = None
                 ):
        """
        :param name: The name of the zone
        :param description: The description of the zone.
        :param zone_type: The type of the zone.
        :param hvac_type: The HVAC type of the zone. Defaults to HVACType.NONE if zone_type is not HVAC.
        """
        self._UID = str(uuid4())
        self._description = description
        self._name = None
        self._zone_type = None
        self._hvac_type = None
        self._adjacent_zones: List['Zone'] = []
        self._overlapping_zones: List['Zone'] = []
        self._spaces: List['AbstractSpace'] = []

        # Apply validation
        self.name = name
        self.zone_type = zone_type
        if zone_type == ZoneType.HVAC:
            self.hvac_type = hvac_type

    @property
    def UID(self) -> str:
        return self._UID

    @property
    def description(self) -> str:
        return self._description

    @description.setter
    def description(self, value: str):
        self._description = value

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, value: str):
        if value is None:
            raise ValueError('name must be a string')
        self._name = value

    @property
    def zone_type(self) -> ZoneType:
        return self._zone_type

    @zone_type.setter
    def zone_type(self, value: ZoneType):
        if value is None:
            raise ValueError('zone_type must be of type ZoneType')
        self._zone_type = value

    @property
    def hvac_type(self) -> HVACType:
        return self._hvac_type

    @hvac_type.setter
    def hvac_type(self, value: ZoneType):
        if self.zone_type != ZoneType.HVAC:
            raise ValueError("HVAC type is only applicable for zones with ZoneType.HVAC.")
        self._hvac_type = value

    def add_adjacent_zones(self, adjacent_zones: List['Zone']):
        """
        adds zones that are adjacent with the current zone.
        :param adjacent_zones: A list of zones that are adjacent to the current zone.
        """
        EntityInsert.insert_building_entity(self._adjacent_zones, adjacent_zones, BuildingEntity.ADJACENT_ZONE.value, self)

    def remove_overlapping_zone(self, overlapping_zone: 'Zone'):
        """
        Removes overlapping zones
        :param overlapping_zone: the overlapping zone to remove
        :return:
        """
        EntityRemover.remove_building_entity(self._overlapping_zones, overlapping_zone,
                                             BuildingEntity.OVERLAPPING_ZONE.value, self)

    def remove_adjacent_zone(self, adjacent_zone: 'Zone'):
        """
        Removes adjacent zones
        :param adjacent_zone: the adjacent zone to remove
        :return:
        """
        EntityRemover.remove_building_entity(self._adjacent_zones, adjacent_zone,
                                             BuildingEntity.ADJACENT_ZONE.value, self)

    def remove_space(self, space: AbstractSpace):
        """
        Removes a space: floor, room, open space from a zone
        :param space: the space to remove
        :return:
        """
        EntityRemover.remove_building_entity(self._spaces, space)

    def add_spaces(self, spaces: List['AbstractSpace']):
        """
        Adds floors or rooms to a zone
        :param spaces: the floors and/or rooms (Abstract spaces) to be added
        """
        EntityInsert.insert_building_entity(self._spaces, spaces,
                                            BuildingEntity.FLOOR_SPACE.value)

    def add_overlapping_zones(self, overlapping_zones: List['Zone']):
        """
        adds zones that overlap with the current zone.
        :param overlapping_zones: A list of zones that are overlapping with the current zone.
        """
        EntityInsert.insert_building_entity(self._overlapping_zones, overlapping_zones,
                                            BuildingEntity.OVERLAPPING_ZONE.value, self)

    def get_adjacent_zone_by_name(self, name) -> 'Zone':
        """
        Search adjacent zones by name
        :param name:  the name of the zone
        :return:
        """
        return StructureEntitySearch.search_by_name(self._adjacent_zones, name)

    def get_adjacent_zone_by_uid(self, uid) -> 'Zone':
        """
        Search adjacent zones by uid
        :param uid: the unique identifier of the adjacent zone
        :return:
        """
        return StructureEntitySearch.search_by_id(self._adjacent_zones, uid)

    def get_adjacent_zones(self, search_terms: Dict = None) -> ['Zone']:
        """
        Search adjacent zones by attributes values
        :param search_terms: a dictionary of attributes and their values
        :return:
        """
        return StructureEntitySearch.search(self._adjacent_zones, search_terms)

    def get_overlapping_zone_by_name(self, name) -> 'Zone':
        """
        Search overlapping zones by name
        :param name:  the name of the zone
        :return:
        """
        return StructureEntitySearch.search_by_name(self._overlapping_zones, name)

    def get_overlapping_zone_by_uid(self, uid) -> 'Zone':
        """
        Search overlapping zones by uid
        :param uid: the unique identifier of the overlapping zone
        :return:
        """
        return StructureEntitySearch.search_by_id(self._overlapping_zones, uid)

    def get_overlapping_zones(self, search_terms: Dict = None) -> ['Zone']:
        """
        Search overlapping zones by attributes values
        :param search_terms: a dictionary of attributes and their values
        :return:
        """
        return StructureEntitySearch.search(self._overlapping_zones, search_terms)

    def get_space_by_uid(self, uid) -> 'AbstractSpace':
        """
        Search spaces by uid
        :param uid: the unique identifier of the space
        :return:
        """
        return StructureSearch.search_by_id(self._spaces, uid)

    def get_spaces(self, search_terms: Dict = None) -> ['AbstractSpace']:
        """
        Search spaces by attributes values
        :param search_terms: a dictionary of attributes and their values
        :return:
        """
        return StructureSearch.search(self._spaces, search_terms)

    def __eq__(self, other):
        # zones are equal if they share the same name
        if isinstance(other, Zone):
            # Check for equality based on the 'name' attribute
            return self.name == other.name
        return False

    def __str__(self):

        zone_details = (
            f"Zone("
            f"UID: {self.UID}, "
            f"Name: {self.name}, "
            f"Description: {self.description}, "
            f"ZoneType: {self.zone_type.value}, "
            f"HVACType: {self.hvac_type.value if self.hvac_type is not None else HVACType.NONE}, "
            f"Adjacent Zones Count: {len(self._adjacent_zones)}, "
            f"Overlapping Zones Count: {len(self._overlapping_zones)})"
        )

        overlapping_zones = "\n".join(str(zone) for zone in self._overlapping_zones)
        spaces = "\n".join(str(space) for space in self._spaces)
        adjacent_zones = "\n".join(str(zone) for zone in self._adjacent_zones)
        return (
            f"{zone_details}\nOverlapping Zones:\n{overlapping_zones}\n"
            f"Adjacent Zones:\n {adjacent_zones}\nSpaces: \n {spaces}"
        )



from metamenth.enumerations.abstract_enum import AbstractEnum


class BuildingType(AbstractEnum):
    """
    Different types of buildings.

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    COMMERCIAL = "Commercial"
    NON_COMMERCIAL = "NonCommercial"
    RESIDENTIAL = "Residential"




from typing import Dict
from metamenth.enumerations import SensorMeasure
from metamenth.enumerations import MeasurementUnit
from datetime import datetime
from metamenth.enumerations import EngineType
from metamenth.enumerations import EngineSubType
from typing import List
from typing import Any
from metamenth.enumerations import RoomType


class Validate:
    """
    Has miscillineous methods for validation

    """

    @staticmethod
    def validate_what3word(input_string: str) -> str:
        """
        Validates that a string is delimited by two "." with three words.

        :param input_string: The string to be validated.
        :return: the input string if valid else raises an error
        """

        if input_string is None or "":
            return ""
        # Split the string using "." as a delimiter
        parts = input_string.split(".")

        # Check if there are exactly three parts
        if len(parts) == 3 and all(part.strip() for part in parts):
            return input_string
        else:
            raise ValueError("Location should be a string of three words delimited with two periods.")

    @staticmethod
    def validate_number_range(value: float, number_range: tuple) -> float:
        try:
            if value is not None:
                if number_range[0] <= value <= number_range[1]:
                    return value
                else:
                    raise ValueError(f"{value} must be a number between {number_range[0]} and {number_range[1]}.")
        except IndexError:
            print(f"{number_range} must be a tuple with just two values")

    @staticmethod
    def validate_none(attributes: Dict):
        none_variables = ""

        for attribute_name, attribute_type in attributes.items():
            if attribute_type is None:
                none_variables = none_variables + attribute_name + " "

        if none_variables:
            raise ValueError("{0} is/are mandatory".format(none_variables.rstrip()))

    @staticmethod
    def parse_date(date_string):
        """
        Returns datetime in the format YYYY-MM-DD HH:MM:SS
        :param date_string: the data string
        :return:
        """
        formats = [
            '%Y-%m-%d %H:%M',
            '%Y-%m-%d',
            '%Y/%m/%d %H:%M:%S.%f',
            '%y/%m/%d %H:%M:%S.%f',
            '%Y/%m/%d %H:%M:%S',
            '%Y/%m/%d %H:%M',
            '%Y/%m/%d',
            '%m/%d/%Y %H:%M',
            '%Y-%m-%d %H:%M:%S.%f',
            '%Y-%m-%d %H:%M:%S'
        ]
        for fmt in formats:
            try:
                dt = datetime.strptime(date_string, fmt)
                return dt.replace(microsecond=0)  # Truncate milliseconds
            except ValueError as err:
                pass
        raise ValueError("No valid date format found")

    @staticmethod
    def validate_sensor_type(sensor_measure: str, unit: str) -> bool:
        """
        Validates the unit of measurement and the type of sensor
        e.g., a temperature sensor should have degree celsius as the measurement unit
        :param sensor_measure: the type of sensor e.g., temperature sensor
        :param unit: the unit of measurement for the type of sensor
        :return: true if the correct unit is used for a specific sensor else false
        Always returns true if sensor measure is other
        """
        if sensor_measure == SensorMeasure.TEMPERATURE.value:
            if unit == MeasurementUnit.DEGREE_CELSIUS.value:
                return True
        elif sensor_measure == SensorMeasure.PRESSURE.value:
            if unit == MeasurementUnit.PASCAL.value:
                return True
        elif sensor_measure == SensorMeasure.CARBON_DIOXIDE.value:
            if unit == MeasurementUnit.PARTS_PER_MILLION.value:
                return True
        elif sensor_measure == SensorMeasure.AIR_VOLUME.value:
            if unit in [MeasurementUnit.LITER.value, MeasurementUnit.CUBIC_FEET.value,
                        MeasurementUnit.CUBIC_CENTIMETER.value, MeasurementUnit.CUBIC_METER.value]:
                return True
        elif sensor_measure in [SensorMeasure.GAS_VELOCITY.value, SensorMeasure.LIQUID_VELOCITY.value]:
            if unit in [MeasurementUnit.METERS_PER_SECOND.value, MeasurementUnit.FEET_PER_SECOND.value]:
                return True
        elif sensor_measure == SensorMeasure.DAYLIGHT.value:
            if unit == MeasurementUnit.LUX.value:
                return True
        elif sensor_measure in [SensorMeasure.DIRECT_RADIATION.value, SensorMeasure.GLOBAL_RADIATION.value]:
            if unit == MeasurementUnit.WATTS_PER_METER_SQUARE.value:
                return True
        elif sensor_measure == SensorMeasure.LUMINANCE.value:
            if unit in [MeasurementUnit.CANDELA_PER_SQUARE_METER.value or MeasurementUnit.NITS.value]:
                return True
        elif sensor_measure == SensorMeasure.NOISE.value:
            if unit == MeasurementUnit.DECIBELS.value:
                return True
        elif sensor_measure == SensorMeasure.OCCUPANCY.value:
            if unit == MeasurementUnit.PRESENCE.value:
                return True
        elif sensor_measure == SensorMeasure.SMOKE.value:
            if unit == MeasurementUnit.MICROGRAM_PER_CUBIC_METER.value:
                return True
        elif sensor_measure == SensorMeasure.CURRENT.value:
            if unit == MeasurementUnit.AMPERE.value:
                return True
        elif sensor_measure == SensorMeasure.VOLTAGE.value:
            if unit == MeasurementUnit.VOLT.value:
                return True
        elif sensor_measure == SensorMeasure.HUMIDITY.value:
            if unit == MeasurementUnit.RELATIVE_HUMIDITY.value:
                return True
        elif sensor_measure == SensorMeasure.OTHER.value:
            return True
        return False

    @staticmethod
    def validate_engine_fuel(engine_type: str, engine_sub_type: str) -> bool:
        """
        Validates engine type and fuel used by such engine types
        """
        if engine_type == EngineType.FUEL_CELL.value:
            if engine_sub_type in [EngineSubType.HYDROGEN.value]:
                return True
        elif engine_type == EngineType.STIRLING.value:
            if engine_sub_type in [EngineSubType.NATURAL_GAS.value, EngineSubType.PROPANE.value,
                                   EngineSubType.BIO_DIESEL.value, EngineSubType.GEO_THERMAL.value,
                                   EngineSubType.ETHANOL.value]:
                return True
        elif engine_type == EngineType.INTERNAL_COMBUSTION.value:
            if engine_sub_type in [EngineSubType.GASOLINE.value, EngineSubType.NATURAL_GAS.value,
                                   EngineSubType.ETHANOL.value, EngineSubType.DIESEL.value,
                                   EngineSubType.BIO_DIESEL.value, EngineSubType.JETFUEL.value]:
                return True
        elif engine_type == EngineType.STEAM.value:
            if engine_sub_type in [EngineSubType.OIL.value, EngineSubType.PEAT.value, EngineSubType.WOOD.value,
                                   EngineSubType.CHARCOAL.value, EngineSubType.COAL.value,
                                   EngineSubType.NATURAL_GAS.value, EngineSubType.BIOMASS.value,
                                   EngineSubType.GEO_THERMAL.value]:
                return True
        elif engine_type == EngineType.ELECTROLYSER.value:
            if engine_sub_type is None:
                return True

        elif engine_type == EngineType.MICRO_TURBINE.value:
            if engine_sub_type in [EngineSubType.NATURAL_GAS.value, EngineSubType.DIESEL.value,
                                   EngineSubType.PROPANE.value, EngineSubType.BIO_DIESEL, EngineSubType.HYDROGEN.value,
                                   EngineSubType.JETFUEL.value]:
                return True
        return False

    @staticmethod
    def is_hvac_component_allowed_in_space(hvac_component, disallowed_entities: List[Any], space_entity):
        """
        Validates HVAC entities that can be added to spaces
        :param hvac_component: the HVAC entity
        :param disallowed_entities: the entities not allowed to be added
        :param space_entity: the space entity, e.g., room, open space
        """
        from metamenth.structure.room import Room
        from metamenth.structure.open_space import OpenSpace
        from metamenth.subsystem.hvac_components.air_volume_box import AirVolumeBox
        from metamenth.subsystem.radiant_slab import RadiantSlab
        from metamenth.subsystem.baseboard_heater import BaseboardHeater
        from metamenth.subsystem.hvac_components.duct import Duct
        from metamenth.subsystem.hvac_components.fan_coil_unit import FanCoilUnit

        if any(isinstance(hvac_component, cls) for cls in disallowed_entities):
            raise ValueError(f'{hvac_component.name} cannot be added to a space entity')
        elif isinstance(space_entity, Room):
            if (space_entity.room_type is not RoomType.MECHANICAL and
                    not any(isinstance(hvac_component, cls) for cls in [AirVolumeBox, BaseboardHeater, RadiantSlab, Duct])
                    and not (isinstance(hvac_component, FanCoilUnit) and not hvac_component.is_ducted)):
                raise ValueError('You can only add HVAC components to mechanical rooms')
        elif (isinstance(space_entity, OpenSpace) and
              not any(isinstance(hvac_component, cls) for cls in [AirVolumeBox, BaseboardHeater, RadiantSlab, Duct])):
            raise ValueError('You can only add air volume box/ducts to open spaces')
        return True



from metamenth.subsystem.hvac_components.interfaces.abstract_duct_connected_component import AbstractDuctConnectedComponent
from metamenth.enumerations import ChillerType
from metamenth.enumerations import PowerState


class Chiller(AbstractDuctConnectedComponent):
    def __init__(self, name: str, chiller_type: ChillerType, power_state: PowerState):
        """
        Models a chiller in a built environment
        :param name: the unique name of the heat exchanger
        :param chiller_type: the type of chiller
        :param power_state: the power state of the chiller
        """
        super().__init__(name)
        self._chiller_type = None
        self._power_state = None

        self.chiller_type = chiller_type
        self.power_state = power_state

    @property
    def chiller_type(self) -> ChillerType:
        return self._chiller_type

    @chiller_type.setter
    def chiller_type(self, value: ChillerType):
        if not value:
            raise ValueError("chiller_type must be of type ChillerType")
        self._chiller_type = value

    @property
    def power_state(self) -> PowerState:
        return self._power_state

    @power_state.setter
    def power_state(self, value: PowerState):
        if value is not None:
            self._power_state = value
        else:
            raise ValueError("power_state must be of type PowerState")

    def __str__(self):
        return (
            f"Chiller ({super().__str__()}"
            f"Type: {self.chiller_type.value}, "
            f"Power State: {self.power_state})"
        )



from metamenth.enumerations.abstract_enum import AbstractEnum


class RefrigerantType(AbstractEnum):
    R134A = "R-134a"
    R410A = "R-410A"
    R22 = "R-22"
    R404A = "R-404A"
    R1234YF = "R-1234yf"
    R1234ZE = "R-1234ze"




from metamenth.subsystem.hvac_components.interfaces.abstract_hvac_component import AbstractHVACComponent
from metamenth.enumerations import MeasurementUnit


class VariableFrequencyDrive(AbstractHVACComponent):
    def __init__(self, name: str, motor_speed: float = 0.0, max_motor_speed: float = 1000.0,
                 speed_measure: MeasurementUnit = MeasurementUnit.REVOLUTIONS_PER_MINUTE):
        """
        Models a boiler in an hvac system
        :param name: the unique name of the boiler
        :param motor_speed: the speed of the VFD
        :param max_motor_speed: the maximum motor speed of the VFD
        :param speed_measure: the measure unit for the motor speed
        :
        """
        super().__init__(name)
        self._motor_speed = motor_speed
        self._max_motor_speed = max_motor_speed
        self._speed_measure = speed_measure

    @property
    def motor_speed(self) -> float:
        return self._motor_speed

    @property
    def max_motor_speed(self) -> float:
        return self._max_motor_speed

    @property
    def speed_measure(self) -> MeasurementUnit:
        return self._speed_measure

    @motor_speed.setter
    def motor_speed(self, value: float):
        if 0 < value <= self._max_motor_speed:
            self._motor_speed = value
        else:
            raise ValueError("motor_speed must be a number greater than 0 and lesser than the maximum motor speed")

    @max_motor_speed.setter
    def max_motor_speed(self, value: float):
        if 0 < value >= self._motor_speed:
            self._max_motor_speed = value
        else:
            raise ValueError("max_motor_speed must be a number greater than 0 and greater than or equal motor speed")

    @speed_measure.setter
    def speed_measure(self, value: MeasurementUnit):
        if value is not None:
            self._speed_measure = value
        else:
            raise ValueError("speed_measure must be of type MeasurementUnit")

    def __str__(self):
        return (
            f"Variable Frequency Drive ({super().__str__()}"
            f"Motor Speed: {self.motor_speed}, "
            f"Maximum Motor Speed: {self._max_motor_speed}, "
            f"Speed Measure: {self.speed_measure})"
        )



from metamenth.enumerations.abstract_enum import AbstractEnum


class CapacitorTech(AbstractEnum):
    """
    Technology used in super capacitors

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    LEAD_BATTERY = "LeadBattery"
    LITHIUM_IRON = "LithiumIron"
    NICKEL = "Nickel"


from metamenth.datatypes.point import Point
from metamenth.datatypes.binary_measure import BinaryMeasure
from metamenth.misc import Validate


class Address:
    """
    Address of a building

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    def __init__(self, city: str, street: str, state: str, zip_code: str, country: str,
                 geocoordinate: Point = None,  what3word: str = None, north_orientation: BinaryMeasure = None):
        """
        :param city: The name of the city
        :param street: The street e.g., 6911 Ave De la Pepiniere.
        :param state: The state (or province) e.g., QC or Quebec.
        :param zip_code: The ZIP code e.g., H1N 1B9
        :param country: the country where the building is located
        :param what3word: The What3word address of the location.
        :param geocoordinate: The geographical coordinates of the location (an instance of Point).
        :param north_orientation: The north orientation of a building.
        """
        self.city = city
        self.street = street
        self.state = state
        self.zip_code = zip_code
        self.country = country
        self.what3word = Validate.validate_what3word(what3word)
        self.geocoordinate = geocoordinate
        self.north_orientation = north_orientation

    def __str__(self):
        geo_coordinate_str = (
            f", GeoCoordinate: (Latitude: {self.geocoordinate.latitude}, Longitude: {self.geocoordinate.longitude})"
            if self.geocoordinate is not None
            else None
        )

        north_orientation_str = (
            f", North Orientation: {self.north_orientation.value}"
            if self.north_orientation is not None
            else None
        )

        return (
            f"GeoLocation("
            f"City: {self.city}, "
            f"Street: {self.street}, "
            f"State: {self.state}, "
            f"ZIP Code: {self.zip_code}, "
            f"Country: {self.country}, "
            f"What3word: {self.what3word}"
            f"{geo_coordinate_str}"
            f"{north_orientation_str})"
        )




from metamenth.enumerations import MeterType
from metamenth.enumerations import MeasurementUnit
from metamenth.enumerations import MeterMeasureMode
from metamenth.measure_instruments.meter_measure import MeterMeasure
from metamenth.enumerations import MeterAccumulationFrequency
from typing import Dict
from metamenth.utils import StructureEntitySearch
from metamenth.measure_instruments.interfaces.abstract_reader import AbstractReader


class Meter(AbstractReader):
    """
    A representation of a meter

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """

    def __init__(self, meter_location: str, measurement_frequency: float, measurement_unit: MeasurementUnit,
                 meter_type: MeterType, measure_mode: MeterMeasureMode, data_accumulated: bool = False,
                 accumulation_frequency: MeterAccumulationFrequency = MeterAccumulationFrequency.NONE,
                 manufacturer: str = None):
        """
        Initializes a Meter instance.

        :param meter_location: The what3word location of the meter.
        :param manufacturer: The manufacturer of the meter.
        :param measurement_frequency: The measurement frequency of the meter.
        :param measurement_unit: The measurement unit of the meter data.
        :param meter_type: The type of the meter.
        :param measure_mode: the data measure mode: manual or automatic
        :param data_accumulated: indicate whether the data is accumulate or not
        :param accumulation_frequency: the frequency at which data is accumulated
        """
        super().__init__(measurement_unit, meter_location, manufacturer)
        self._measurement_frequency = None
        self._meter_type = None
        self._meter_measures: [MeterMeasure] = []
        self._measure_mode = None
        self._data_accumulated = data_accumulated
        self._accumulation_frequency = MeterAccumulationFrequency.NONE

        # Apply validation
        self.measurement_frequency = measurement_frequency
        self.meter_type = meter_type
        self.measure_mode = measure_mode
        self.accumulation_frequency = accumulation_frequency

    @property
    def measurement_frequency(self) -> float:
        return self._measurement_frequency

    @measurement_frequency.setter
    def measurement_frequency(self, value: float):
        if value is not None:
            self._measurement_frequency = value
        else:
            raise ValueError("measurement_frequency must be a float")

    @property
    def measure_mode(self) -> MeterMeasureMode:
        return self._measure_mode

    @measure_mode.setter
    def measure_mode(self, value: MeterMeasureMode):
        if value is not None:
            self._measure_mode = value
        else:
            raise ValueError("measure_mode must be a float")

    @property
    def data_accumulated(self) -> bool:
        return self._data_accumulated

    @data_accumulated.setter
    def data_accumulated(self, value: bool):
        if value is not None:
            self._data_accumulated = value
        else:
            raise ValueError("data_accumulated must be a boolean")

    @property
    def accumulation_frequency(self) -> MeterAccumulationFrequency:
        return self._accumulation_frequency

    @accumulation_frequency.setter
    def accumulation_frequency(self, value: MeterAccumulationFrequency):
        if value is not None:
            if self.data_accumulated and value is None:
                raise ValueError("accumulation_frequency must not be None")
            else:
                self._accumulation_frequency = value
        else:
            raise ValueError("data_accumulated must be a boolean")

    @property
    def meter_type(self) -> MeterType:
        return self._meter_type

    @meter_type.setter
    def meter_type(self, value: MeterType):
        if value is not None:
            self._meter_type = value
        else:
            raise ValueError("Meter type must be of type MeterType")

    def get_meter_measures(self, search_terms: Dict = None) -> [MeterMeasure]:
        """
        Search meter recordings by attributes values
        :param search_terms: a dictionary of attributes and their values
        :return [MeterMeasure]:
        """
        return StructureEntitySearch.search(self._meter_measures, search_terms)

    def get_meter_measure_by_date(self, from_timestamp: str, to_timestamp: str = None) -> [MeterMeasure]:
        """
        searches meter recordings based on provided timestamp
        :param from_timestamp: the start timestamp
        :param to_timestamp: the end timestamp
        :return: [MeterMeasure]
        """
        return StructureEntitySearch.date_range_search(self._meter_measures, from_timestamp, to_timestamp)

    def add_meter_measure(self, meter_measure: MeterMeasure):
        """
        Add measurement for this meter
        :param meter_measure: the recorded measurement by the meter.
        """
        self._meter_measures.append(meter_measure)

    def __str__(self):
        """
        :return: A formatted string representing the meter.
        """
        measurements = "\n".join(str(measure) for measure in self.get_meter_measures())

        return (
            f"Meter("
            f"{super().__str__()} "
            f"Frequency: {self.measurement_frequency}, "
            f"Measure Mode: {self.measure_mode.value}, "
            f"Data Accumulated: {self.data_accumulated}, "
            f"Accumulation Frequency: {self.accumulation_frequency.value}, "
            f"Type: {self.meter_type.value}, "
            f"Measurements: {measurements})"
        )



from metamenth.enumerations import MeasurementUnit
from metamenth.datatypes.binary_measure import BinaryMeasure
from metamenth.energysystem.interfaces.abstract_common_energy_system import AbstractCommonEnergySystem


class AbstractEnergySystem(AbstractCommonEnergySystem):
    def __init__(self, name: str, inverter: bool, unit: MeasurementUnit):
        super().__init__(name)
        self._inverter = None
        self._unit = None
        self._manufacturing_year = None
        self._capacity = None

        self.inverter = inverter
        self.unit = unit

    @property
    def inverter(self) -> bool:
        return self._inverter

    @inverter.setter
    def inverter(self, value: bool):
        if value is None:
            raise ValueError("inverter should be of type bool")
        self._inverter = value

    @property
    def unit(self) -> MeasurementUnit:
        return self._unit

    @unit.setter
    def unit(self, value: MeasurementUnit):
        if value is None:
            raise ValueError("unit should be of type MeasurementUnit")
        self._unit = value

    @property
    def manufacturing_year(self) -> int:
        return self._manufacturing_year

    @manufacturing_year.setter
    def manufacturing_year(self, value: int):
        if value is None:
            raise ValueError("manufacturing_year should be of type int")
        self._manufacturing_year = value

    @property
    def capacity(self) -> BinaryMeasure:
        return self._capacity

    @capacity.setter
    def capacity(self, value: BinaryMeasure):
        self._capacity = value

    def __str__(self):
        return (
            f"{super().__str__()}"
            f"Inverter: {self.inverter}, "
            f"Unit: {self.unit}, "
            f"Capacity: {self.capacity}, "
            f"Manufacturing Year: {self.manufacturing_year}, "
        )



from metamenth.structure.interfaces.abstract_space import AbstractSpace
from metamenth.datatypes.interfaces.abstract_measure import AbstractMeasure
from metamenth.enumerations import FloorType
from metamenth.structure.open_space import OpenSpace
from metamenth.structure.room import Room
from metamenth.utils import EntityRemover
from metamenth.enumerations import BuildingEntity
from metamenth.utils import EntityInsert
from typing import Union
from metamenth.utils import StructureSearch
from typing import Dict
from typing import List


class Floor(AbstractSpace):
    """
    A floor on a building

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """

    def __init__(
        self,
        area: AbstractMeasure,
        number: int,
        floor_type: FloorType,
        height: AbstractMeasure = None,
        description: str = None,
        open_spaces: [OpenSpace] = None,
        rooms: [Room] = None,
        location: str = None
    ):
        """
        :param area (BinaryMeasure): The area of the floor.
        :param location: The location of the floor (three words terminated with a period).
        :param description: A description of the floor.
        :param number: The floor number.
        :param floor_type: The type of floor (enum).
        :param open_spaces: Initial open spaces(s) on floor.
        :param rooms: Initial room(s) on floor.
        """
        super().__init__(area, location)
        self._description = description
        self._height = height
        self._number = None
        self._floor_type = None
        self._open_spaces: List['OpenSpace'] = []
        self._rooms: List['Room'] = []

        # apply validation
        self.number = number
        self.floor_type = floor_type

        if open_spaces:
            self.add_open_spaces(open_spaces)

        if rooms:
            self.add_rooms(rooms)

        # A floor should have at least one open space or one room
        if not self._open_spaces and not self._rooms:
            raise ValueError("A floor must have at least one room or one open space.")

    @property
    def description(self) -> str:
        return self._description

    @description.setter
    def description(self, value: str):
        self._description = value

    @property
    def number(self) -> int:
        return self._number

    @number.setter
    def number(self, value: int):
        if value is not None:
            self._number = value
        else:
            raise ValueError("number must be an int")

    @property
    def floor_type(self) -> FloorType:
        return self._floor_type

    @floor_type.setter
    def floor_type(self, value: FloorType):
        if value is not None:
            self._floor_type = value
        else:
            raise ValueError("floor_type must be of type FloorType")

    @property
    def height(self) -> AbstractMeasure:
        return self._height

    @height.setter
    def height(self, value: AbstractMeasure):
        if value is not None:
            self._height = value
        else:
            raise ValueError("height must be of type AbstractMeasure")

    @property
    def open_spaces(self):
        raise AttributeError("Cannot get open_spaces")

    @open_spaces.setter
    def open_spaces(self, value: Union[OpenSpace, List[OpenSpace]]):
        if value is not None:
            if isinstance(value, list):
                self._open_spaces = value
            else:
                self._open_spaces.append(value)
        else:
            raise ValueError("open_spaces must be of type [OpenSpace]")

    @property
    def rooms(self):
        raise AttributeError("Cannot get rooms")

    @rooms.setter
    def rooms(self, value: Union[Room, List[Room]]):
        if value is not None:
            if isinstance(value, list):
                self._rooms = value
            else:
                self._rooms.append(value)
        else:
            raise ValueError("rooms must be of type [Room]")

    def add_open_spaces(self, open_spaces: List['OpenSpace']):
        """
        Add one or multiple OpenSpaces to the floor.
        :param open_spaces: The open spaces to add to the floor.
        """
        EntityInsert.insert_building_entity(self._open_spaces, open_spaces, BuildingEntity.FLOOR_SPACE.value)
        return self

    def add_rooms(self, rooms: List['Room']):
        """
        Add one or multiple rooms to the floor.
        :param rooms: The open spaces to add to the floor.
        """
        EntityInsert.insert_building_entity(self._rooms, rooms, BuildingEntity.FLOOR_SPACE.value)
        return self

    def remove_open_space(self, open_space: OpenSpace):
        """
        Removes open space from a floor
        :param open_space: the open space entity to remove
        :return:
        """
        EntityRemover.remove_building_entity(self._open_spaces, open_space, BuildingEntity.OPEN_SPACE.value, self)
        return self

    def remove_room(self, room: Room):
        """
        Removes room from a floor
        :param room: the room to remove
        :return:
        """
        EntityRemover.remove_building_entity(self._rooms, room, BuildingEntity.ROOM.value, self)

    def get_open_space_by_uid(self, uid: str) -> OpenSpace:
        """
        Retrieves an open space given the uid
        :param uid: the uid of the open space
        :return:
        """
        return StructureSearch.search_by_id(self._open_spaces, uid)

    def get_room_by_uid(self, uid: str) -> Room:
        """
        Retrieves a room given the uid
        :param uid: the uid of the room
        :return:
        """
        return StructureSearch.search_by_id(self._rooms, uid)

    def get_open_space_by_name(self, name: str) -> OpenSpace:
        """
        Retrieves an open space given the name
        :param name: the name of the open space
        :return:
        """
        return StructureSearch.search_by_name(self._open_spaces, name)

    def get_room_by_name(self, name: str) -> Room:
        """
        Retrieves a room given its name
        :param name: the name of the room
        :return:
        """
        return StructureSearch.search_by_name(self._rooms, name)

    def get_rooms(self, search_term: Dict = None) -> List[Room]:
        """
        Retrieves rooms that match attributes and their values
        :param search_term: attributes and their values
        :return:
        """
        return StructureSearch.search(self._rooms, search_term)

    def get_open_spaces(self, search_term: Dict = None) -> List[OpenSpace]:
        """
        Retrieves open spaces that match attributes and their values
        :param search_term: attributes and their values
        :return:
        """
        return StructureSearch.search(self._open_spaces, search_term)

    def accept(self, visitor):
        """
        visitor method to accept
        visit operation for the current floor
        :param visitor: the visitor object
        """
        visitor.visit_floor(self)

    def __eq__(self, other):
        # floors are equal if they share the same number
        if isinstance(other, Floor):
            # Check for equality based on the 'number' attribute
            return self.number == other.number
        return False

    def __str__(self):
        floor_details = (f"Floor {super().__str__()} {self.number} ({self.floor_type.value}): {self.description}, "
                         f"Area: {self.area}, Height: {self.height}, Location: {self.location}, UID: {self.UID}, "
                         f"Rooms Count: {len(self._rooms)}, Open Spaces Count: {len(self._open_spaces)})")

        rooms = "\n".join(str(room) for room in self._rooms)
        open_spaces = "\n".join(str(space) for space in self._open_spaces)

        return f"{floor_details}\nRooms:\n{rooms}\nOpen Space:\n {open_spaces})"



from metamenth.enumerations.abstract_enum import AbstractEnum


class EngineType(AbstractEnum):
    """
    Types of Engine

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    FUEL_CELL = "FuelCell"
    STIRLING = "Stirling"
    INTERNAL_COMBUSTION = "InternalCombustion"
    MICRO_TURBINE = "MicroTurbine"
    ELECTROLYSER = "Electrolyser"
    STEAM = "Steam"
    OTHER = "Other"



from metamenth.enumerations.abstract_enum import AbstractEnum


class PowerState(AbstractEnum):
    """
    Power states of an HVAC component

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    ON = "On"
    OFF = "Off"
    STANDBY = "StandBy"
    OUT_OF_SERVICE = "OutOfService"
    VARIABLE_SPEED = "VariableSpeed"
    INTERMITTENT = "Intermittent"
    BOOST = "Boost"
    NIGHT_MODE = "NightMode"
    NONE = "None"
    OTHER = "Other"




from metamenth.datatypes.interfaces.abstract_range_measure import AbstractRangeMeasure
from metamenth.datatypes.measure import Measure


class LinearMeasure(AbstractRangeMeasure):
    """
    Linear measurement

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    def __init__(self, measure: Measure):
        """
        :param measure: the measurement object
        """
        super().__init__(measure)
        self.slope = measure.slope

    def __str__(self):
        return (
            f"LinearMeasure("
            f"Slope: {self.slope}, "
            f"{super().__str__()})"
        )




from metamenth.enumerations import BuildingEntity


class EntityRemover:
    """
    A visitor that removes entities into other entities

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """

    def __init__(self):
        pass

    @staticmethod
    def remove_building_entity(building_entity_list, entity, entity_type=None, entity_object=None):
        """
        Removes a building entity: floor, meter, weather station and schedule
        :param entity: the entity to be removed
        :param entity_type: a string representing the entity (e.g., floor) to remove
        :param building_entity_list: the building whose entity is being removed
        :param entity_object: the object whose list has an item being removed
        :return:
        """

        if entity_type == BuildingEntity.ZONE.value:
            # then remove the zone from the list of zones for the space
            entity.remove_space(entity_object)
            building_entity_list.remove(entity)
        else:
            building_entity_list.remove(entity)




from metamenth.measure_instruments.interfaces.abstract_data_measure import AbstractDataMeasure


class DamperPosition(AbstractDataMeasure):

    def __init__(self, value: float, timestamp: str = None):
        """
        :param value: The numerical value measured
        :param timestamp: the time of measurement
        """
        super().__init__(value, timestamp)

